[Version = "1.3.0"]
section AutodeskDataConnector;

// Forge OAuth2 values
client_id_us = "dmEANmD9HlGhP5ZTZj83eqnFioVQyKor";
client_id_emea = "3pKYkRFebEuGeTeJatHkDkTgZfY2kMjt";

redirect_uri = "https://oauth.powerbi.com/views/oauthredirect.html";

base_path = "https://developer.api.autodesk.com";
prefix = "/data-connector/v1/accounts/";
accounts_prefix = "/bim360/";
authorize_uri = base_path & "/authentication/v2/authorize";
scope = "data:read data:write data:create account:read";
region = "";

getClientIdByRegion = (region) => 
    let
        clientId = if region = "EU" then client_id_emea else client_id_us
    in
        clientId;

// ======================= PKCE =========================================

StartLogin = (resourceUrl, state, display) =>
    let
        toJson = Json.Document(resourceUrl),
        clientId = getClientIdByRegion(toJson[region]),
        // We'll generate our code verifier using Guids
        codeVerifier = Text.NewGuid() & Text.NewGuid(),
        AuthorizeUrl = authorize_uri & "?" & Uri.BuildQueryString([
            client_id = clientId,
            response_type = "code",
            code_challenge_method = "plain",
            scope="account:read data:create data:read data:write",
            code_challenge = codeVerifier,
            state = state,
            redirect_uri = redirect_uri])
    in
        [
            LoginUri = AuthorizeUrl,
            CallbackUri = redirect_uri,
            WindowHeight = 720,
            WindowWidth = 1024,
            // Need to roundtrip this value to FinishLogin
            Context = codeVerifier
        ];

// The code verifier will be passed in through the context parameter.
FinishLogin = (c, dataSourcePath, context, callbackUri, state) =>
    let
        Parts = Uri.Parts(callbackUri)[Query]
    in
        TokenMethod(dataSourcePath, Parts[code], "authorization_code", context);

// Verifier is optional to support both the original FinishLogin call
// (which has a verifier) and the Refresh call (which does not).
TokenMethod = (dataSourcePath, code, grant_type, optional verifier) =>
    let
        toJson = Json.Document(dataSourcePath),
        clientId = getClientIdByRegion(toJson[region]),
        codeVerifier = if (verifier <> null) then [code_verifier = verifier] else [],
        codeParameter = if (grant_type = "authorization_code") then [ code = code ] else [ refresh_token = code ],
        query = codeVerifier & codeParameter & [
            client_id = clientId,
            grant_type = grant_type,
            redirect_uri = redirect_uri
        ],
 
        // Set this if your API returns a non-2xx status for login failures
        // ManualHandlingStatusCodes = {400, 403}
        ManualHandlingStatusCodes= {},
        
        Response = Web.Contents(base_path & "/authentication/v2" & "/token", [
            Content = Text.ToBinary(Uri.BuildQueryString(query)),
            Headers = [
                #"Content-type" = "application/x-www-form-urlencoded",
                #"Accept" = "application/json"
            ],
            ManualStatusHandling = ManualHandlingStatusCodes
        ]),
        Parts = Json.Document(Response)
    in
        // check for error in response
        if (Parts[error]? <> null) then 
            error Error.Record(Parts[error], Parts[message]?)
        else
            Parts;
 
Refresh = (resourceUrl, refresh_token) => TokenMethod(refresh_token, "refresh_token");



// ===================================================================


// Data Source Kind description
AutodeskDataConnector = [
    TestConnection = (header) => { "AutodeskDataConnector.Contents", header },
    Authentication = [
        OAuth = [
            StartLogin = StartLogin,
            FinishLogin = FinishLogin,
            Refresh = Refresh
        ]
    ],
    Label = Extension.LoadString("DataSourceLabel")
];

[DataSource.Kind="AutodeskDataConnector", Publish="AutodeskDataConnector.Publish"]

shared AutodeskDataConnector.Contents = Value.ReplaceType(AutodeskDataConnectorImpl, AutodeskDataConnectorType);

getHeader = (region as text) => 
    let 
        headerRegion = if (region = "US") then "US" else "EMEA"
    in 
        [
            Origin = "PowerBI Desktop",
            #"x-ads-region" = headerRegion
        ];       


AutodeskDataConnectorType = type function (
    region as (type text meta [
        Documentation.FieldCaption = "Region",
        // Documentation.FieldDescription = "",
        Documentation.AllowedValues = { "US", "EU" }
    ])) 
    as table meta [
        Documentation.Name = "Autodesk Data Connector - Preview",
        Documentation.LongDescription = "Version: " & Extension.LoadString("Version")
     ];

AutodeskDataConnectorImpl = (
     region as text
    ) =>
    let
        header = getHeader(region),
        accounts =  getAccounts(header),
        accountsNavigation = generateAccountsNavigation(accounts, header)
    in
        accountsNavigation;

getAccounts = (header) =>
    let
        response = Json.Document(
            Web.Contents(base_path & accounts_prefix & "admin/v1/accounts",
                [
                    Query = [ limit="200", fields="name"],
                    Headers = header
                ]
            )
        ),
        accounts = response[results]
    in
        accounts;

/**
 * Generate a NavigationTable for the accounts
 */
 
generateAccountsNavigation = (accounts, header) =>
    let
        #"Converted to Table" = Table.FromList(accounts, Splitter.SplitByNothing(), null, null, ExtraValues.Error),
        #"Expanded Column1" = Table.ExpandRecordColumn(#"Converted to Table", "Column1", {"name", "id"}, {"Name", "Key"}),
        #"Added Data" = Table.AddColumn(
            #"Expanded Column1", "Data", each generateFileTables([Key], header)
        ),
        #"Added ItemKind" = Table.AddColumn(#"Added Data", "ItemKind", each "Folder", Text.Type),
        #"Added ItemName" = Table.AddColumn(#"Added ItemKind", "ItemName", each "Folder", Text.Type),
        #"Added IsLeaf" = Table.AddColumn(#"Added ItemName", "IsLeaf", each false, Logical.Type),
        asNav = Table.ToNavigationTable(#"Added IsLeaf", {"Name"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        asNav;

generateFileTables = (id, header) =>
    let
        values = getViews(id, header),
        #"Added Data" = Table.AddColumn(
            values[data], "Data", each getCsv(values[account], values[job], [Name], header)
        ),
        #"Added ItemKind" = Table.AddColumn(#"Added Data", "ItemKind", each "Table", Text.Type),
        #"Added ItemName" = Table.AddColumn(#"Added ItemKind", "ItemName", each "Table", Text.Type),
        #"Added IsLeaf" = Table.AddColumn(#"Added ItemName", "IsLeaf", each true, Text.Type),
        asNav = Table.ToNavigationTable(#"Added IsLeaf", {"Name"}, "Name", "Data", "ItemKind", "ItemName", "IsLeaf")
    in
        asNav;

/**
 * Request
 */


 
fetchRequests = (accountId, header) => 
        let 
            waitForResult = Value.WaitFor(
                (iteration) =>
                    let 
                        result =  Web.Contents(
                            base_path & prefix & accountId  & "/requests", 
                            [
                                Query = [ sort="desc", limit="10", offset="0" ],
                                Headers = header
                            ]
                        ),
                        buffered = Binary.Buffer(result),
                        status = Value.Metadata(result)[Response.Status],
                        resultParse = Json.Document(result),
                        actualResult = if status = 502 then null else Json.Document(result)
                    in
                        actualResult,
                    (iteration) => #duration(0, 0, 0, Number.Power(2, iteration)),
            5)
        in
            waitForResult;


AutodeskDataConnector.getLastRequest = (requests) =>
    try
        let
            results = requests[results],
            #"Converted to Table" = Table.FromList(results, Splitter.SplitByNothing(), null, null, ExtraValues.Error),
            #"Expanded Column1" = Table.ExpandRecordColumn(
                #"Converted to Table", "Column1", { "id", "lastQueuedAt" }, { "id", "lastQueuedAt" }
            ),
            getFirstAndQueued = Table.Sort(
                #"Expanded Column1",
                {"lastQueuedAt", Order.Descending}
            ),
            request = getFirstAndQueued{0}
        in
            request
    otherwise
        error Error.Record("Not found", "No data requests for this account in the last 30 days. Users with executive overview access can request data from Autodesk Data Connector-Preview.");

/**
 * Jobs
 */
 

fetchJobs = (accountId, requestId, header) => 
    let 
        waitForResult = Value.WaitFor(
            (iteration) =>
                let 
                    result =  Web.Contents(
                        base_path & prefix & accountId & "/requests/" & requestId & "/jobs",
                            [
                                Query = [ sort="desc", limit="100", offset="0" ],
                                Headers = header
                            ]
                    ),
                    buffered = Binary.Buffer(result),
                    status = Value.Metadata(result)[Response.Status],
                    resultParse = Json.Document(result),
                    actualResult = if status = 502 then null else Json.Document(result)
                in
                    actualResult,
                (iteration) => #duration(0, 0, 0, Number.Power(2, iteration)),
        5)
    in
        waitForResult;


AutodeskDataConnector.getLastJob = (jobs) =>
    try
        let

            jobResults = jobs[results],
            jobTable = Table.FromList(
                jobResults, Splitter.SplitByNothing(), null, null, ExtraValues.Error
            ),
            #"Expanded Job Column" = Table.ExpandRecordColumn(
                jobTable, "Column1", { "id" }, { "id" }
            ),
            jobId = #"Expanded Job Column"{0}
        in
            jobId
    otherwise
        error Error.Record("Not found", "No data requests for this account in the last 30 days. Users with executive overview access can request data from Autodesk Data Connector-Preview.");

/**
 * Data listing
 */

fetchDataListing = (accountId, jobId, header) => 
    try
        let 
            waitForResult = Value.WaitFor(
                (iteration) =>
                    let 
                        result =  Web.Contents(
                            base_path & prefix & accountId & "/jobs/" & jobId & "/data-listing",
                            [
                                Headers = header
                            ]
                        ),
                        buffered = Binary.Buffer(result),
                        status = Value.Metadata(result)[Response.Status],
                        resultParse = Json.Document(result),
                        actualResult = if status = 502 then null else Json.Document(result)
                    in
                        actualResult,
                    (iteration) => #duration(0, 0, 0, Number.Power(2, iteration)),
            5)
        in
            waitForResult
    otherwise
        error Error.Record("Error", "Unfortunately, your Data Connector job failed. Please re-submit a new Data Connector request to continue.");

   
AutodeskDataConnector.prepareDataListing = (dataListing) =>
    let
        dataListToTable = Table.FromList(dataListing, Splitter.SplitByNothing(), null, null, ExtraValues.Error),
        #"Expanded Datalist Column" = Table.ExpandRecordColumn(
            dataListToTable, 
            "Column1", 
            {"name"},
            {"Name"}
        ),
        #"Filtered Rows" = Table.SelectRows(
            #"Expanded Datalist Column", 
            each Text.Contains([Name], ".csv")
        ),
        dataListTable = Table.ReplaceValue(
            #"Filtered Rows", ".csv", "", Replacer.ReplaceText, {"Name"}
        )
    in
        dataListTable;

/**
 * CSV
 */


fetchSignedUrl = (accountId, jobId, filename, header) => 
    let 
        waitForResult = Value.WaitFor(
            (iteration) =>
                let 
                    result =  Web.Contents(
                        base_path & prefix & accountId & "/jobs/" & jobId & "/data/" & filename & ".csv", [
                            Headers = header,
                            ManualStatusHandling = {502}
                        ]),
                    buffered = Binary.Buffer(result),
                    status = Value.Metadata(result)[Response.Status],
                    resultParse = Json.Document(result),
                    actualResult = if status = 502 then null else resultParse[signedUrl]
                in
                    actualResult,
            (iteration) => #duration(0, 0, 0, Number.Power(2, iteration)),
        5)
    in
        waitForResult;

fetchCsvFile = (accountId, jobId, filename, header) =>
    let
        waitForResult = Value.WaitFor(
            (iteration) =>
                let
                    signedUrl = fetchSignedUrl(accountId, jobId, filename, header),
                    result = Web.Contents(signedUrl, [ 
                        ManualCredentials = true,
                        ManualStatusHandling = {403}
                    ]),
                    buffered = Binary.Buffer(result),
                    status = Value.Metadata(result)[Response.Status],
                    actualResult = if 
                            status = 403
                        then null else Csv.Document(buffered, [ Delimiter=",", Encoding=1252, QuoteStyle=QuoteStyle.Csv])
                in
                    actualResult,
            (iteration) => #duration(0, 0, 0, Number.Power(2, iteration)),
        5)
    in
        waitForResult;

getCsv = (accountId, jobId, name, header) =>
    let
        csv = fetchCsvFile(accountId, jobId, name, header),
        withHeaders = Table.PromoteHeaders(csv)
    in
        withHeaders;

/**
 * Get Datalisting
 */
 
getViews = (accountId as text, header) =>
    try
        let
            requests        = fetchRequests(accountId, header),
            requestId       = AutodeskDataConnector.getLastRequest(requests)[id],
            jobs            = fetchJobs(accountId, requestId, header),
            jobId           = AutodeskDataConnector.getLastJob(jobs)[id],
            dataListing     = fetchDataListing(accountId, jobId, header),
            values          = [
                data = AutodeskDataConnector.prepareDataListing(dataListing),
                account = accountId,
                job = jobId
            ]
        in
            values
    otherwise
        error Error.Record(
            "Not found", 
            "No data requests for this account in the last 30 days. Users with executive overview access can request data from Autodesk Data Connector-Preview."
        );
/**
 * Helper functions for OAuth2: StartLogin, FinishLogin
 */


/**
 * Common functions
 */

Value.WaitFor = (producer as function, interval as function, optional count as number) as any =>
    let
        list = List.Generate(
            () => {0, null},
            (state) => state{0} <> null and (count = null or state{0} < count),
            (state) => if state{1} <> null then {null, state{1}} else {1 + state{0}, Function.InvokeAfter(() => producer(state{0}), interval(state{0}))},
            (state) => state{1})
    in
        List.Last(list);

Table.ToNavigationTable = (
    table as table,
    keyColumns as list,
    nameColumn as text,
    dataColumn as text,
    itemKindColumn as text,
    itemNameColumn as text,
    isLeafColumn as text
) as table =>
    let
        tableType = Value.Type(table),
        newTableType = Type.AddTableKey(tableType, keyColumns, true) meta
        [
            NavigationTable.NameColumn = nameColumn,
            NavigationTable.DataColumn = dataColumn,
            NavigationTable.ItemKindColumn = itemKindColumn,
            Preview.DelayColumn = itemNameColumn,
            NavigationTable.IsLeafColumn = isLeafColumn
        ],
        navigationTable = Value.ReplaceType(table, newTableType)
    in
        navigationTable;

/**
 * Data Source UI publishing description
 */
AutodeskDataConnector.Publish = [
    Beta = false,
    Category = "Other",
    ButtonText = { "Autodesk Data Connector - Preview", "Autodesk Data Connector - Preview" },
    LearnMoreUrl = "https://help.autodesk.com/view/BIM360D/ENU/?guid=BIM360D_Insight_data_extractor_html",
    SourceImage = AutodeskDataConnector.Icons,
    SourceTypeImage = AutodeskDataConnector.Icons
];

AutodeskDataConnector.Icons = [
    Icon16 = { Extension.Contents("AutodeskDataConnector16.png"), Extension.Contents("AutodeskDataConnector20.png"), Extension.Contents("AutodeskDataConnector24.png"), Extension.Contents("AutodeskDataConnector32.png") },
    Icon32 = { Extension.Contents("AutodeskDataConnector32.png"), Extension.Contents("AutodeskDataConnector40.png"), Extension.Contents("AutodeskDataConnector48.png"), Extension.Contents("AutodeskDataConnector64.png") }
];
